@@ -5,7 +5,7 @@
 # Written 91-12-02 through 92-01-01 by Stephen McGee.
 # Modified 92-02-11 through 92-02-22 by Chris Arthur to further generalize.
 #
-# Copyright (C) 1991-2020 Free Software Foundation, Inc.
+# Copyright (C) 1991-2022 Free Software Foundation, Inc.
 # This file is part of GNU Make.
 #
 # GNU Make is free software; you can redistribute it and/or modify it under
@@ -33,6 +33,7 @@
 use Config;
 use Cwd;
 use File::Spec;
+use File::Temp;
 
 # The number of test categories we've run
 $categories_run = 0;
@@ -47,6 +48,10 @@ $tests_run = 0;
 # The number of tests in this category that have passed
 $tests_passed = 0;
 
+$port_type = undef;
+$osname = undef;
+$vos = undef;
+$pathsep = undef;
 
 # Yeesh.  This whole test environment is such a hack!
 $test_passed = 1;
@@ -55,40 +60,53 @@ $test_passed = 1;
 $test_timeout = 5;
 $test_timeout = 10 if $^O eq 'VMS';
 
+$diff_name = undef;
+
+# Create a temporary directory that tests can use, outside the temp
+# directory that make is using.
+$TEMPDIR = File::Temp->newdir();
+
 # Path to Perl
 $perl_name = $^X;
 if ($^O ne 'VMS') {
     $perl_name .= $Config{_exe} unless $perl_name =~ m/$Config{_exe}$/i;
 }
-# If it's a simple name, look it up on PATH
-{
-    my ($v,$d,$f) = File::Spec->splitpath($perl_name);
-    if (!$d) {
-        my $perl = undef;
-        foreach my $p (File::Spec->path()) {
-            my $f = File::Spec->catfile($p, $f);
-            if (-e $f) {
-                $perl = $f;
-                last;
-            }
-        }
-        if ($perl) {
-            $perl_name = $perl;
-        } else {
-            print "Cannot locate Perl interpreter $perl_name\n";
-        }
+
+sub which {
+  my $cmd = $_[0];
+
+  # Poor man's File::Which
+  my ($v,$d,$f) = File::Spec->splitpath($cmd);
+  if ($d) {
+    # The command has a pathname so don't look for it in PATH.
+    # Use forward-slashes even on Windows, else it fails in recipes.
+    (-f $cmd and -x _) or return undef;
+    $cmd =~ tr,\\,/,;
+    return $cmd;
+  }
+
+  my @ext;
+  if ($port_type eq 'UNIX' || $port_type eq 'VMS-DCL') {
+    @ext = ('');
+  } else {
+    @ext = index($f, '.') == -1 ? () : ('');
+    push @ext, split /;/, $ENV{PATHEXT};
+  }
+
+  foreach my $dir (File::Spec->path()) {
+    foreach my $e (@ext) {
+      my $p = File::Spec->catfile($dir, "$cmd$e");
+      (-f $p and -x _) or next;
+      # Use forward-slashes even on Windows, else it fails in recipes.
+      $p =~ tr,\\,/,;
+      return $p;
     }
+  }
+  return undef;
 }
-# Make sure it uses forward-slashes even on Windows, else it won't work
-# in recipes
-$perl_name =~ tr,\\,/,;
 
-# %makeENV is the cleaned-out environment.
-%makeENV = ();
-
-# %extraENV are any extra environment variables the tests might want to set.
-# These are RESET AFTER EVERY TEST!
-%extraENV = ();
+# %makeENV is the cleaned-out environment.  Tests must not modify it.
+my %makeENV = ();
 
 sub vms_get_process_logicals {
   # Sorry for the long note here, but to keep this test running on
@@ -150,27 +168,45 @@ sub resetENV
       }
     }
   }
+}
 
-  foreach $v (keys %extraENV) {
-    $ENV{$v} = $extraENV{$v};
-    delete $extraENV{$v};
-  }
+# Returns a string-ified version of cmd which is a value provided to exec()
+# so it can either be a ref of a list or a string.
+sub cmd2str
+{
+    my $cmd = $_[0];
+    if (!ref($cmd)) {
+        return $cmd;
+    }
+
+    my @c;
+    foreach (@$cmd) {
+        if (/[][#;"*?&|<>(){}\$`^~!]/) {
+            s/\'/\'\\'\'/g;
+            push @c, "'$_'";
+        } else {
+            push @c, $_;
+        }
+    }
+    return join(' ', @c);
 }
 
 sub toplevel
 {
+  %origENV = %ENV unless $^O eq 'VMS';
+
   # Pull in benign variables from the user's environment
 
-  foreach (# UNIX-specific things
+  foreach (# POSIX-specific things
            'TZ', 'TMPDIR', 'HOME', 'USER', 'LOGNAME', 'PATH',
            'LD_LIBRARY_PATH',
-           # SAN things
+           # *SAN things
            'ASAN_OPTIONS', 'UBSAN_OPTIONS',
            # Purify things
            'PURIFYOPTIONS',
-           # Windows NT-specific stuff
-           'Path', 'SystemRoot',
-           # DJGPP-specific stuff
+           # Windows-specific things
+           'Path', 'SystemRoot', 'TEMP', 'TMP', 'USERPROFILE', 'PATHEXT',
+           # DJGPP-specific things
            'DJDIR', 'DJGPP', 'SHELL', 'COMSPEC', 'HOSTNAME', 'LFN',
            'FNCASE', '387', 'EMU387', 'GROUP'
           ) {
@@ -180,11 +216,10 @@ sub toplevel
   # Make sure our compares are not foiled by locale differences
 
   $makeENV{LC_ALL} = 'C';
+  $makeENV{LANG} = 'C';
+  $makeENV{LANGUAGE} = 'C';
 
   # Replace the environment with the new one
-  #
-  %origENV = %ENV unless $^O eq 'VMS';
-
   resetENV();
 
   $| = 1;                     # unbuffered output
@@ -195,6 +230,7 @@ sub toplevel
   $detail = 0;                # detailed verbosity
   $keep = 0;                  # keep temp files around
   $workdir = "work";          # The directory where the test will start running
+  $tempdir = "_tmp";          # A temporary directory
   $scriptdir = "scripts";     # The directory where we find the test scripts
   $tmpfilesuffix = "t";       # the suffix used on tmpfiles
   $default_output_stack_level = 0;  # used by attach_default_output, etc.
@@ -204,12 +240,37 @@ sub toplevel
 
   &get_osname;  # sets $osname, $vos, $pathsep, and $short_filenames
 
+  $perl_name = which($perl_name);
+
+  # See if we have a diff
+  $diff_name = which('diff');
+  if (!$diff_name) {
+      print "No diff found; differences will not be shown\n";
+  }
+
   &set_defaults;  # suite-defined
 
   &parse_command_line (@ARGV);
 
   print "OS name = '$osname'\n" if $debug;
 
+  $temppath = File::Spec->rel2abs($tempdir);
+
+  if (-d $temppath) {
+    print "Clearing $temppath...\n";
+    &remove_directory_tree("$temppath/")
+      or &error ("Couldn't wipe out $temppath: $!\n");
+  } else {
+    mkdir ($temppath, 0777) or error ("Cannot mkdir $temppath: $!\n");
+  }
+
+  # This is used by POSIX systems
+  $makeENV{TMPDIR} = $temppath;
+
+  # These are used on Windows
+  $makeENV{TMP} = $temppath;
+  $makeENV{TEMP} = $temppath;
+
   $workpath = "$cwdslash$workdir";
   $scriptpath = "$cwdslash$scriptdir";
 
@@ -237,9 +298,9 @@ sub toplevel
   if (-d $workpath) {
     print "Clearing $workpath...\n";
     &remove_directory_tree("$workpath/")
-        or &error ("Couldn't wipe out $workpath\n");
+      or &error ("Couldn't wipe out $workpath: $!\n");
   } else {
-    mkdir ($workpath, 0777) or &error ("Couldn't mkdir $workpath: $!\n");
+    mkdir ($workpath, 0777) or &error ("Cannot mkdir $workpath: $!\n");
   }
 
   if (!-d $scriptpath) {
@@ -253,8 +314,8 @@ sub toplevel
         $dir = $1;
         push (@rmdirs, $dir);
         -d "$workpath/$dir"
-           or mkdir ("$workpath/$dir", 0777)
-           or &error ("Couldn't mkdir $workpath/$dir: $!\n");
+            or mkdir ("$workpath/$dir", 0777)
+            or &error ("Couldn't mkdir $workpath/$dir: $!\n");
       }
     }
   } else {
@@ -266,7 +327,7 @@ sub toplevel
     foreach my $dir (@dirs) {
       next if ($dir =~ /^(\..*|CVS|RCS)$/ || ! -d "$scriptpath/$dir");
       push (@rmdirs, $dir);
-      # VMS can have overlayed file systems, so directories may repeat.
+      # VMS can have overlaid file systems, so directories may repeat.
       next if -d "$workpath/$dir";
       mkdir ("$workpath/$dir", 0777)
           or &error ("Couldn't mkdir $workpath/$dir: $!\n");
@@ -293,6 +354,8 @@ sub toplevel
     rmdir ("$workpath/$dir");
   }
 
+  rmdir ($temppath);
+
   $| = 1;
 
   $categories_failed = $categories_run - $categories_passed;
@@ -319,6 +382,8 @@ sub get_osname
 {
   # Set up an initial value.  In perl5 we can do it the easy way.
   $osname = defined($^O) ? $^O : '';
+  $vos = 0;
+  $pathsep = "/";
 
   # find the type of the port.  We do this up front to have a single
   # point of change if it needs to be tweaked.
@@ -354,10 +419,7 @@ sub get_osname
     $port_type = 'UNIX';
   }
 
-  if ($osname eq 'VMS')
-  {
-    $vos = 0;
-    $pathsep = "/";
+  if ($osname eq 'VMS') {
     return;
   }
 
@@ -387,7 +449,7 @@ sub get_osname
     $vos = 1;
     $pathsep = ">";
 
-  } else {
+  } elsif ($osname eq '') {
     # the following is regrettably gnarly, but it seems to be the only way
     # to not get ugly error messages if uname can't be found.
     # Hmmm, BSD/OS 2.0's uname -a is excessively verbose.  Let's try it
@@ -402,8 +464,6 @@ sub get_osname
         $osname = "(something posixy)";
       }
     }
-    $vos = 0;
-    $pathsep = "/";
   }
 
   if (! $short_filenames) {
@@ -497,12 +557,12 @@ sub print_centered
 sub print_banner
 {
   # $testee is suite-defined
-  my $info = "Running tests for $testee on $osname\n";
-  my $len = &max (length($info), length($testee_version), 73) + 5;
+  my $info = "Running tests for $testee on $osname";
+  my $len = &max (length($info), length($testee_version), 77) + 2;
   my $line = ("-" x $len) . "\n";
 
   &print_centered ($len, $line);
-  &print_centered ($len, $info);
+  &print_centered ($len, $info."\n");
   &print_centered ($len, $testee_version);
   &print_centered ($len, $line);
   print "\n";
@@ -515,11 +575,31 @@ sub run_all_tests
 
   $categories_run = 0;
 
+  # Make a copy of STDIN so we can reset it
+  open(INCOPY, "<&STDIN");
+
+  # Leave enough space in the extensions to append a number, even
+  # though it needs to fit into 8+3 limits.
+  if ($short_filenames) {
+    $logext = 'l';
+    $diffext = 'd';
+    $baseext = 'b';
+    $runext = 'r';
+    $extext = '';
+  } else {
+    $logext = 'log';
+    $diffext = 'diff';
+    $baseext = 'base';
+    $runext = 'run';
+    $extext = '.';
+  }
+
   $lasttest = '';
   # $testname is published
   foreach $testname (sort @TESTS) {
     # Skip duplicates on VMS caused by logical name search lists.
     next if $testname eq $lasttest;
+
     $lasttest = $testname;
     $suite_passed = 1;       # reset by test on failure
     $num_of_logfiles = 0;
@@ -531,21 +611,6 @@ sub run_all_tests
     $perl_testname = "$scriptpath$pathsep$testname";
     $testname =~ s/(\.pl|\.perl)$//;
     $testpath = "$workpath$pathsep$testname";
-    # Leave enough space in the extensions to append a number, even
-    # though it needs to fit into 8+3 limits.
-    if ($short_filenames) {
-      $logext = 'l';
-      $diffext = 'd';
-      $baseext = 'b';
-      $runext = 'r';
-      $extext = '';
-    } else {
-      $logext = 'log';
-      $diffext = 'diff';
-      $baseext = 'base';
-      $runext = 'run';
-      $extext = '.';
-    }
     $extext = '_' if $^O eq 'VMS';
     $log_filename = "$testpath.$logext";
     $diff_filename = "$testpath.$diffext";
@@ -559,7 +624,7 @@ sub run_all_tests
 
     $output = "........................................................ ";
 
-    substr($output,0,length($testname)) = "$testname ";
+    substr($output, 0, length($testname)) = "$testname ";
 
     print $output;
 
@@ -569,6 +634,9 @@ sub run_all_tests
     # Run the test!
     $code = do $perl_testname;
 
+    # Reset STDIN from the copy in case it was changed
+    open(STDIN, "<&INCOPY");
+
     ++$categories_run;
     $total_tests_run += $tests_run;
     $total_tests_passed += $tests_passed;
@@ -641,6 +709,8 @@ sub run_all_tests
 
     print "$status\n";
   }
+
+  close(INCOPY);
 }
 
 # If the keep flag is not set, this subroutine deletes all filenames that
@@ -740,13 +810,29 @@ sub error
   die "$caller: $message";
 }
 
+my %old_tempfiles = ();
+
 sub compare_output
 {
-  my ($answer,$logfile) = @_;
-  my ($slurp, $answer_matched) = ('', 0);
+  my ($answer, $logfile) = @_;
+  my ($slurp, $answer_matched, $extra) = ('', 0, 0);
 
   ++$tests_run;
 
+  my @tf = ();
+  foreach my $file (glob(File::Spec->catfile($temppath, "*"))) {
+    if (!exists $old_tempfiles{$file}) {
+      push @tf, $file;
+      $old_tempfiles{$file} = 1;
+    }
+  }
+  if (@tf) {
+    open (LOGFILE, '>>', $logfile) or die "Cannot open log file $logfile: $!\n";
+    print LOGFILE "Leftover temporary files: @tf\n";
+    close (LOGFILE);
+    $extra = 1;
+  }
+
   if (! defined $answer) {
     print "Ignoring output ........ " if $debug;
     $answer_matched = 1;
@@ -899,7 +985,12 @@ sub compare_output
     }
   }
 
-  if ($answer_matched && $test_passed) {
+  if ($keep || ! $answer_matched) {
+    &create_file(&get_basefile, $answer);
+    &create_file(&get_runfile, $command_string);
+  }
+
+  if ($answer_matched && $test_passed && !$extra) {
     print "ok\n" if $debug;
     ++$tests_passed;
     return 1;
@@ -908,15 +999,16 @@ sub compare_output
   if (! $answer_matched) {
     print "DIFFERENT OUTPUT\n" if $debug;
 
-    &create_file (&get_basefile, $answer);
-    &create_file (&get_runfile, $command_string);
-
     print "\nCreating Difference File ...\n" if $debug;
 
     # Create the difference file
-
-    my $command = "diff -c " . &get_basefile . " " . $logfile;
-    &run_command_with_output(&get_difffile,$command);
+    my $base = get_basefile();
+    if ($diff_name) {
+        my $command = "$diff_name -c $base $logfile";
+        &run_command_with_output(get_difffile(), $command);
+    } else {
+        create_file(get_difffile(), "Log file $logfile differs from base file $base\n");
+    }
   }
 
   return 0;
@@ -947,7 +1039,7 @@ sub attach_default_output
   if ($vos)
   {
     my $code = system "++attach_default_output_hack $filename";
-    $code == -2 or &error ("adoh death\n", 1);
+    $code == -2 or &error ("ado death\n", 1);
     return 1;
   }
 
@@ -1059,15 +1151,13 @@ sub _run_with_timeout
 # This runs a command without any debugging info.
 sub _run_command
 {
-  # We reset this before every invocation.  On Windows I think there is only
-  # one environment, not one per process, so I think that variables set in
-  # test scripts might leak into subsequent tests if this isn't reset--???
-  resetENV();
-
   my $orig = $SIG{ALRM};
   my $code = eval { _run_with_timeout(@_); };
   $SIG{ALRM} = $orig;
 
+  # Reset then environment so that it's clean for the next test.
+  resetENV();
+
   if ($@) {
     # The eval failed.  If it wasn't SIGALRM then die.
     $@ eq "timeout\n" or die "Command failed: $@";
@@ -1129,8 +1219,9 @@ sub remove_directory_tree
   -e $targetdir or return 1;
 
   &remove_directory_tree_inner ("RDT00", $targetdir) or return 0;
-  if ($nuketop) {
-    rmdir($targetdir) or return 0;
+  if ($nuketop && !rmdir ($targetdir)) {
+    print "Cannot remove $targetdir: $!\n";
+    return 0;
   }
 
   return 1;
@@ -1149,10 +1240,16 @@ sub remove_directory_tree_inner
 
     lstat ($object);
     if (-d _ && &remove_directory_tree_inner ($subdirhandle, $object)) {
-      rmdir $object or return 0;
+      if (!rmdir($object)) {
+        print "Cannot remove $object: $!\n";
+        return 0;
+      }
     } else {
       if ($^O ne 'VMS') {
-        unlink $object or return 0;
+        if (!unlink $object) {
+          print "Cannot unlink $object: $!\n";
+          return 0;
+        }
       } else {
         # VMS can have multiple versions of a file.
         1 while unlink $object;
